use std::sync::{Arc, LazyLock, RwLock, atomic::AtomicBool};

use cpal::traits::{DeviceTrait, HostTrait, StreamTrait};
use protocol::{
    Error,
    protocol::{AUDIO_DATA_SENDER_CHANNEL_LENGTH, Signal, SignalType},
};
use tokio::sync::{broadcast, mpsc};

use crate::gui::State;

const LATENCY_THRESHOLD: usize = 1024 * 8;

struct SpeakerSink {
    speaker_id: u8,
    audio_data_sender: broadcast::Sender<f32>,
    audio_data_receiver: RwLock<broadcast::Receiver<f32>>,
}

impl SpeakerSink {
    fn new(speaker_id: u8) -> Self {
        let (audio_data_sender, audio_data_receiver) =
            broadcast::channel(AUDIO_DATA_SENDER_CHANNEL_LENGTH);
        let audio_data_receiver = audio_data_receiver.into();
        Self {
            speaker_id,
            audio_data_sender,
            audio_data_receiver,
        }
    }
}

pub async fn record(
    mut record_control: mpsc::Receiver<State>,
    record_sender: Arc<broadcast::Sender<f32>>,
) -> Result<(), Error> {
    let host = cpal::default_host();
    let input_device = host
        .default_input_device()
        .ok_or(Error::Record("Input Device".to_string()))?;
    let config = input_device
        .default_input_config()
        .map_err(|inner| Error::Record(inner.to_string()))?
        .into();
    println!("Recorder Stream Config = {:#?}", config);

    let input = move |data: &[f32], _: &cpal::InputCallbackInfo| {
        for &sample in data {
            if record_sender.receiver_count() > 0 && sample != 0.0 {
                if let Err(err_val) = record_sender.send(sample) {
                    eprintln!("Error: Microphone Send | {}", err_val);
                }
            }
        }
    };

    let input_stream = input_device
        .build_input_stream(&config, input, voice_error, None)
        .map_err(|inner| Error::Record(inner.to_string()))?;

    input_stream
        .pause()
        .map_err(|inner| Error::Record(inner.to_string()))?;

    tokio::task::block_in_place(|| {
        loop {
            match record_control.blocking_recv() {
                Some(message) => match message {
                    State::Active => input_stream
                        .play()
                        .map_err(|inner| Error::Record(inner.to_string()))?,
                    State::Passive => input_stream
                        .pause()
                        .map_err(|inner| Error::Record(inner.to_string()))?,
                    State::Loading => {}
                },
                None => {
                    input_stream
                        .pause()
                        .map_err(|inner| Error::Record(inner.to_string()))?;
                    return Ok(());
                }
            }
        }
    })
}

static ONLINE_SPEAKERS: LazyLock<RwLock<Vec<SpeakerSink>>> = LazyLock::new(|| vec![].into());

async fn signal_handler(
    mut play_receiver: broadcast::Receiver<Signal>,
    play_pause: Arc<AtomicBool>,
) {
    while let Ok(signal) = play_receiver.recv().await {
        match signal.get_signal_type() {
            SignalType::AudioDatum => {
                let data = signal.get_audio_datum();

                let speaker_list_search_result = ONLINE_SPEAKERS.read().unwrap().binary_search_by(
                    |speaker_sink: &SpeakerSink| {
                        speaker_sink.speaker_id.cmp(&signal.get_speaker_id())
                    },
                );

                match speaker_list_search_result {
                    Ok(speaker_sink_index) => {
                        let online_speakers = ONLINE_SPEAKERS.read().unwrap();
                        let speaker_sink = online_speakers.get(speaker_sink_index).expect("Never");

                        if play_pause.load(std::sync::atomic::Ordering::Relaxed) {
                            if let Err(err_val) = speaker_sink.audio_data_sender.send(data) {
                                eprintln!("Error: Send Audio Datum | Local | {}", err_val);
                            }
                        }
                    }
                    Err(_) => {
                        let speaker_sink = SpeakerSink::new(signal.get_speaker_id());

                        if play_pause.load(std::sync::atomic::Ordering::Relaxed) {
                            if let Err(err_val) = speaker_sink.audio_data_sender.send(data) {
                                eprintln!("Error: Send Audio Datum | Local | {}", err_val);
                            }
                        }

                        let mut online_speakers = ONLINE_SPEAKERS.write().unwrap();
                        online_speakers.push(speaker_sink);
                        online_speakers.sort_by(|x, y| x.speaker_id.cmp(&y.speaker_id));
                    }
                }
            }
            SignalType::SpeakerLeft => {
                ONLINE_SPEAKERS
                    .write()
                    .unwrap()
                    .retain(|speaker_sink| speaker_sink.speaker_id != signal.get_speaker_id());
            }
        }
    }
}

pub async fn play(
    mut play_control: mpsc::Receiver<State>,
    play_receiver: broadcast::Receiver<Signal>,
) -> Result<(), Error> {
    let host = cpal::default_host();
    let output_device = host.default_output_device().unwrap();
    let config = output_device.default_output_config().unwrap().into();
    println!("Speaker Stream Config = {:#?}", config);

    let output = move |data: &mut [f32], _: &cpal::OutputCallbackInfo| {
        for sample in data {
            match ONLINE_SPEAKERS.try_read() {
                Ok(online_speakers) => {
                    *sample = online_speakers
                        .iter()
                        .map(
                            |speaker_sink| match speaker_sink.audio_data_receiver.try_write() {
                                Ok(mut speaker_sink_audio_receiver) => {
                                    if speaker_sink_audio_receiver.len() > LATENCY_THRESHOLD {
                                        println!(
                                            "Warn: Latency Fix Agent | Local | Audio | {}",
                                            speaker_sink_audio_receiver.len()
                                        );
                                        *speaker_sink_audio_receiver =
                                            speaker_sink_audio_receiver.resubscribe();
                                    }
                                    match speaker_sink_audio_receiver.try_recv() {
                                        Ok(audio_datum) => audio_datum,
                                        Err(_) => 0.0,
                                    }
                                }
                                Err(_) => 0.0,
                            },
                        )
                        .sum::<f32>()
                        .clamp(-1.0, 1.0);
                }
                Err(_) => *sample = 0.0,
            }
        }
    };

    let play_pause = Arc::new(AtomicBool::new(true));

    let signal_handler = tokio::spawn(signal_handler(play_receiver, play_pause.clone()));

    let output_stream = output_device
        .build_output_stream(&config, output, voice_error, None)
        .unwrap();

    output_stream
        .play()
        .map_err(|inner| Error::Play(inner.to_string()))?;

    tokio::task::block_in_place(|| {
        loop {
            match play_control.blocking_recv() {
                Some(message) => match message {
                    State::Active => {
                        play_pause.store(true, std::sync::atomic::Ordering::Relaxed);
                        // output_stream
                        //     .play()
                        //     .map_err(|inner| Error::Play(inner.to_string()))?
                    }
                    State::Passive => {
                        play_pause.store(false, std::sync::atomic::Ordering::Relaxed);
                        // I'm not pausing at hardware level to prevent some queue and permission
                        // problems at the first hand

                        // output_stream
                        //     .pause()
                        //     .map_err(|inner| Error::Play(inner.to_string()))?
                    }
                    State::Loading => {}
                },
                None => {
                    output_stream
                        .pause()
                        .map_err(|inner| Error::Play(inner.to_string()))?;

                    signal_handler.abort();
                    return Ok(());
                }
            }
        }
    })
}

fn voice_error(err_val: cpal::StreamError) {
    eprintln!("Error: Voice Error | {}", err_val);
}
