use std::sync::{Arc, RwLock};

use iced::{
    Alignment::Center,
    Element, Task, Theme,
    widget::{button, column, row},
};
use protocol::Error;
use tokio::sync::{
    broadcast::{self},
    mpsc, oneshot,
};

use crate::{
    ClientConfig, MICROPHONE_BUFFER_LENGHT, SPEAKER_BUFFER_LENGHT,
    audio::{play, record},
    stream::{connect, disconnect_watcher},
};

#[derive(Debug)]
struct Controller {
    record_control: mpsc::Sender<State>,
    play_control: mpsc::Sender<State>,
    connection_stop_sender: RwLock<Option<oneshot::Sender<bool>>>,
}

impl Controller {
    fn reset_connection(&self) -> Result<(), Error> {
        let connection_stop_sender = self.connection_stop_sender.write().unwrap().take();
        match connection_stop_sender {
            Some(connection_stop_sender) => {
                connection_stop_sender.send(true).unwrap();

                Ok(())
            }
            None => Err(Error::NotSupposeTo(
                "Connection Stop Sender is None".to_string(),
            ))?,
        }
    }
}

#[derive(Debug)]
struct Channel {
    microphone: Arc<broadcast::Sender<f32>>,
    speaker: Arc<broadcast::Sender<protocol::protocol::Signal>>,
}

#[derive(Debug, Clone, Copy)]
struct GUIStatus {
    room: State,
    microphone: State,
    speaker: State,
}
impl Default for GUIStatus {
    fn default() -> Self {
        Self {
            room: Default::default(),
            microphone: Default::default(),
            speaker: State::Active,
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum State {
    Active,
    Passive,
    Loading,
}
impl Default for State {
    fn default() -> Self {
        State::Passive
    }
}

#[derive(Debug, Clone, Copy)]
pub enum Message {
    None,
    JoinRoom,
    LeaveRoom,
    UnmuteMicrophone,
    MuteMicrophone,
    UnmuteSpeaker,
    MuteSpeaker,
}

#[derive(Debug)]
pub struct App {
    client_config: Arc<ClientConfig>,
    gui_status: Arc<RwLock<GUIStatus>>,
    channel: Channel,
    controller: Arc<Controller>,
}

impl App {
    fn reset_connection(&mut self) -> Result<(), Error> {
        self.controller.reset_connection()?;
        self.gui_status.write().unwrap().room = State::Passive;
        Ok(())
    }

    pub fn theme(&self) -> Theme {
        Theme::Dark
    }

    pub fn new() -> Self {
        let record_sender = Arc::new(broadcast::channel(MICROPHONE_BUFFER_LENGHT).0);

        let record_control = mpsc::channel(1);
        tokio::spawn(record(record_control.1, record_sender.clone()));

        let play_sender = Arc::new(broadcast::channel(SPEAKER_BUFFER_LENGHT).0);

        let play_control = mpsc::channel(1);
        tokio::spawn(play(play_control.1, play_sender.clone().subscribe()));

        App {
            client_config: ClientConfig::new().into(),
            gui_status: RwLock::new(GUIStatus::default()).into(),
            channel: Channel {
                microphone: record_sender,
                speaker: play_sender,
            },
            controller: Controller {
                record_control: record_control.0,
                play_control: play_control.0,
                connection_stop_sender: None.into(),
            }
            .into(),
        }
    }
    pub fn view(&self) -> Element<'_, Message> {
        let join_room_button = match self.gui_status.read().unwrap().room {
            State::Active => button("Leave Room").on_press(Message::LeaveRoom),
            State::Passive => button("Join Room").on_press(Message::JoinRoom),
            State::Loading => button("Loading"),
        };

        let microphone_button = match self.gui_status.read().unwrap().microphone {
            State::Active => button("Microphone ON").on_press(Message::MuteMicrophone),
            State::Passive => button("Microphone OFF").on_press(Message::UnmuteMicrophone),
            State::Loading => button("Microphone Loading"),
        };

        let speaker_button = match self.gui_status.read().unwrap().speaker {
            State::Active => button("Speaker ON").on_press(Message::MuteSpeaker),
            State::Passive => button("Speaker OFF").on_press(Message::UnmuteSpeaker),
            State::Loading => button("Speaker Loading"),
        };

        column![
            row![join_room_button, microphone_button, speaker_button]
                .spacing(20)
                .align_y(Center)
        ]
        .align_x(Center)
        .into()
    }
    pub fn update(&mut self, message: Message) -> Task<Message> {
        match message {
            Message::None => Task::none(),
            Message::JoinRoom => {
                self.gui_status.write().unwrap().room = State::Loading;
                let client_config = self.client_config.clone();
                let gui_status = self.gui_status.clone();
                let microphone_receiver = self.channel.microphone.subscribe();
                let speaker_sender = self.channel.speaker.clone();

                let (connection_stop_sender, connection_stop_receiver) = oneshot::channel();
                *self.controller.connection_stop_sender.write().unwrap() =
                    Some(connection_stop_sender);

                Task::perform(
                    async move {
                        match connect(microphone_receiver, speaker_sender, client_config).await {
                            Ok(connect_return) => {
                                tokio::spawn(disconnect_watcher(
                                    connection_stop_receiver,
                                    connect_return,
                                ));
                                gui_status.write().unwrap().room = State::Active;
                                Some((gui_status, Message::UnmuteSpeaker))
                            }
                            Err(err_val) => {
                                eprintln!("Error: Connect | {}", err_val);
                                gui_status.write().unwrap().room = State::Passive;
                                None
                            }
                        }
                    },
                    |inner_return| match inner_return {
                        Some((gui_status, speaker_activate_message)) => {
                            if gui_status.read().unwrap().speaker == State::Passive {
                                speaker_activate_message
                            } else {
                                Message::None
                            }
                        }
                        None => Message::None,
                    },
                )
            }
            Message::LeaveRoom => {
                self.gui_status.write().unwrap().room = State::Loading;
                if let Err(err_val) = self.reset_connection() {
                    eprintln!("Error: Leave Room | {}", err_val);
                }
                Task::none()
            }
            Message::UnmuteMicrophone => {
                self.gui_status.write().unwrap().microphone = State::Loading;

                let gui_status = self.gui_status.clone();
                let signal = self.controller.clone();
                Task::perform(
                    async move { signal.record_control.send(State::Active).await },
                    move |result| {
                        match result {
                            Ok(_) => {
                                gui_status.write().unwrap().microphone = State::Active;
                            }
                            Err(err_val) => {
                                eprintln!("Error: Unmute Microphone | {}", err_val);
                                gui_status.write().unwrap().microphone = State::Passive;
                            }
                        }
                        Message::None
                    },
                )
            }
            Message::MuteMicrophone => {
                self.gui_status.write().unwrap().microphone = State::Loading;

                let gui_status = self.gui_status.clone();
                let signal = self.controller.clone();
                Task::perform(
                    async move { signal.record_control.send(State::Passive).await },
                    move |result| {
                        match result {
                            Ok(_) => {
                                gui_status.write().unwrap().microphone = State::Passive;
                            }
                            Err(err_val) => {
                                eprintln!("Error: Mute Microphone | {}", err_val);
                                gui_status.write().unwrap().microphone = State::Active;
                            }
                        }
                        Message::None
                    },
                )
            }
            Message::UnmuteSpeaker => {
                self.gui_status.write().unwrap().speaker = State::Loading;

                let gui_status = self.gui_status.clone();
                let signal = self.controller.clone();
                Task::perform(
                    async move { signal.play_control.send(State::Active).await },
                    move |result| {
                        match result {
                            Ok(_) => {
                                gui_status.write().unwrap().speaker = State::Active;
                            }
                            Err(err_val) => {
                                eprintln!("Error: Unmute Speaker | {}", err_val);
                                gui_status.write().unwrap().speaker = State::Passive;
                            }
                        }
                        Message::None
                    },
                )
            }
            Message::MuteSpeaker => {
                self.gui_status.write().unwrap().speaker = State::Loading;

                let gui_status = self.gui_status.clone();
                let signal = self.controller.clone();
                Task::perform(
                    async move { signal.play_control.send(State::Passive).await },
                    move |result| {
                        match result {
                            Ok(_) => {
                                gui_status.write().unwrap().speaker = State::Passive;
                            }
                            Err(err_val) => {
                                eprintln!("Error: Mute Speaker | {}", err_val);
                                gui_status.write().unwrap().speaker = State::Active;
                            }
                        }
                        Message::None
                    },
                )
            }
        }
    }
}
