use std::{net::SocketAddr, path::Path, sync::Arc};

use protocol::{
    Error,
    protocol::{NETWORK_DATA_LENGTH, Signal},
};
use s2n_quic::{
    Client,
    client::Connect,
    stream::{ReceiveStream, SendStream},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    sync::{broadcast, oneshot},
    task::JoinHandle,
};

use crate::ClientConfig;

#[derive(Debug)]
pub struct ConnectReturn {
    send_signals_task: JoinHandle<()>,
    receive_signals_task: JoinHandle<()>,
}

pub async fn connect(
    microphone_receiver: broadcast::Receiver<f32>,
    speaker_sender: Arc<broadcast::Sender<Signal>>,
    client_config: Arc<ClientConfig>,
) -> Result<ConnectReturn, Error> {
    let client = Client::builder()
        .with_io("0:0")
        .map_err(|err_val| Error::ConnectionSetup(err_val.to_string()))?
        .with_tls(Path::new(&client_config.certificate_path))
        .map_err(|err_val| Error::ConnectionSetup(err_val.to_string()))?
        .start()
        .map_err(|err_val| Error::ConnectionSetup(err_val.to_string()))?;

    println!("Client Address = {}", client.local_addr().unwrap());
    let connect = Connect::new(
        client_config
            .server_address
            .parse::<SocketAddr>()
            .map_err(|inner| Error::Connection(inner.to_string()))?,
    )
    .with_server_name("localhost");
    let mut connection = match client.connect(connect).await {
        Ok(connection) => connection,
        Err(err_val) => {
            eprintln!("Error: Client Connection | {}", err_val);
            return Err(Error::Connection(err_val.to_string()));
        }
    };
    connection
        .keep_alive(true)
        .map_err(|err_val| Error::ConnectionSetup(err_val.to_string()))?;

    let stream = connection
        .open_bidirectional_stream()
        .await
        .map_err(|err_val| Error::ConnectionSetup(err_val.to_string()))?;

    let (receive_stream, send_stream) = stream.split();

    let send_signals_task = tokio::spawn(send_signals(send_stream, microphone_receiver));
    let receive_signals_task = tokio::spawn(receive_signals(receive_stream, speaker_sender));

    Ok(ConnectReturn {
        send_signals_task,
        receive_signals_task,
    })
}

pub async fn disconnect_watcher(
    connection_stop_receiver: oneshot::Receiver<bool>,
    connect_return: ConnectReturn,
) {
    if let Err(err_val) = connection_stop_receiver.await {
        eprintln!(
            "Error: Receive Connection Stop Signal | Local | {}",
            err_val
        );
    }

    println!("Going to Disconnect");

    connect_return.send_signals_task.abort();
    connect_return.receive_signals_task.abort();

    println!("Disconnected");
}

#[allow(unused_assignments)]
async fn send_signals(
    mut send_stream: SendStream,
    mut microphone_receiver: broadcast::Receiver<f32>,
) {
    loop {
        match microphone_receiver.recv().await {
            Ok(audio_datum) => {
                if let Err(err_val) = send_stream.write_f32(audio_datum).await {
                    eprintln!("Error: Send Audio Datum | Remote | {}", err_val);
                    return;
                }
            }
            Err(err_val) => {
                eprintln!("Error: Receive Audio Datum | Local | {}", err_val);
                match err_val {
                    broadcast::error::RecvError::Closed => return,
                    broadcast::error::RecvError::Lagged(lag_amount) => {
                        println!("Warn: Latency Fix Agent | Local | Signal | {}", lag_amount);
                        microphone_receiver = microphone_receiver.resubscribe();
                        todo!(
                            "This might be problematic, when we deleted the queue maybe user left and another user joined with same id ?"
                        );
                    }
                }
            }
        }
    }
}

async fn receive_signals(
    mut receive_stream: ReceiveStream,
    speaker_sender: Arc<broadcast::Sender<Signal>>,
) {
    let mut network_buffer = [0; NETWORK_DATA_LENGTH];
    loop {
        match receive_stream.read_exact(&mut network_buffer).await {
            Ok(_) => match Signal::unpack(network_buffer) {
                Ok(signal) => {
                    let _ = speaker_sender.send(signal);
                }
                Err(err_val) => {
                    eprintln!("Error: Unpack Signal | {}", err_val);
                }
            },
            Err(err_val) => {
                eprintln!("Error: Receive Audio Data | Remote | {}", err_val);
                todo!("GUI Status Disconnect");
                // break;
            }
        }
    }
}
